@@ -2,14 +2,14 @@
 # http://www.unidata.ucar.edu/software/netcdf/docs/file_format_specifications.html#netcdf_4_spec
 import os.path
 import warnings
+import weakref
 from collections import ChainMap, OrderedDict, defaultdict
 from collections.abc import Mapping
-from distutils.version import LooseVersion
 
 import h5py
 import numpy as np
+from packaging import version
 
-# from .compat import ChainMap, OrderedDict, unicode
 from .attrs import Attributes
 from .dimensions import Dimensions
 from .utils import Frozen
@@ -19,11 +19,13 @@ try:
 except ImportError:
     no_h5pyd = True
     h5_group_types = (h5py.Group,)
+    h5_dataset_types = (h5py.Dataset,)
 else:
     no_h5pyd = False
     h5_group_types = (h5py.Group, h5pyd.Group)
+    h5_dataset_types = (h5py.Dataset, h5pyd.Dataset)
 
-__version__ = "0.10.0"
+__version__ = "0.13.0"
 
 
 _NC_PROPERTIES = "version=2,h5netcdf=%s,hdf5=%s,h5py=%s" % (
@@ -53,17 +55,8 @@ class CompatibilityError(Exception):
     """Raised when using features that are not part of the NetCDF4 API."""
 
 
-def _invalid_netcdf_feature(feature, allow, file, stacklevel=0):
-    if allow is None:
-        msg = (
-            "{} are supported by h5py, but not part of the NetCDF API. "
-            "You are writing an HDF5 file that is not a valid NetCDF file! "
-            "In the future, this will be an error, unless you set "
-            "invalid_netcdf=True.".format(feature)
-        )
-        warnings.warn(msg, FutureWarning, stacklevel=stacklevel)
-        file._write_ncproperties = False
-    elif not allow:
+def _invalid_netcdf_feature(feature, allow):
+    if not allow:
         msg = (
             "{} are not a supported NetCDF feature, and are not allowed by "
             "h5netcdf unless invalid_netcdf=True.".format(feature)
@@ -73,12 +66,20 @@ def _invalid_netcdf_feature(feature, allow, file, stacklevel=0):
 
 class BaseVariable(object):
     def __init__(self, parent, name, dimensions=None):
-        self._parent = parent
-        self._root = parent._root
+        self._parent_ref = weakref.ref(parent)
+        self._root_ref = weakref.ref(parent._root)
         self._h5path = _join_h5paths(parent.name, name)
         self._dimensions = dimensions
         self._initialized = True
 
+    @property
+    def _parent(self):
+        return self._parent_ref()
+
+    @property
+    def _root(self):
+        return self._root_ref()
+
     @property
     def _h5ds(self):
         # Always refer to the root file and store not h5py object
@@ -87,7 +88,8 @@ class BaseVariable(object):
 
     @property
     def name(self):
-        return self._h5ds.name
+        # fix name if _nc4_non_coord_
+        return self._h5ds.name.replace("_nc4_non_coord_", "")
 
     def _lookup_dimensions(self):
         attrs = self._h5ds.attrs
@@ -97,7 +99,8 @@ class BaseVariable(object):
                 order_dim[coord_id] for coord_id in attrs["_Netcdf4Coordinates"]
             )
 
-        child_name = self.name.split("/")[-1]
+        # need to use the h5ds name here to distinguish from collision dimensions
+        child_name = self._h5ds.name.split("/")[-1]
         if child_name in self._parent.dimensions:
             return (child_name,)
 
@@ -125,6 +128,26 @@ class BaseVariable(object):
             dims.append(name)
         return tuple(dims)
 
+    def _attach_dim_scales(self):
+        """Attach dimension scales"""
+        for n, dim in enumerate(self.dimensions):
+            self._h5ds.dims[n].attach_scale(self._parent._all_h5groups[dim])
+
+    def _attach_coords(self):
+        dims = self.dimensions
+        coord_ids = np.array([self._parent._dim_order[d] for d in dims], "int32")
+        if len(coord_ids) > 1:
+            self._h5ds.attrs["_Netcdf4Coordinates"] = coord_ids
+
+    def _ensure_dim_id(self):
+        """Set _Netcdf4Dimid"""
+        # set _Netcdf4Dimid, use id of first dimension
+        # netCDF4 does this when the first variable's data is written
+        if self.dimensions and not self._h5ds.attrs.get("_Netcdf4Dimid", False):
+            dim = self._parent._all_h5groups[self.dimensions[0]]
+            if "_Netcdf4Dimid" in dim.attrs:
+                self._h5ds.attrs["_Netcdf4Dimid"] = dim.attrs["_Netcdf4Dimid"]
+
     @property
     def dimensions(self):
         if self._dimensions is None:
@@ -206,10 +229,14 @@ class Variable(BaseVariable):
 
 class _LazyObjectLookup(Mapping):
     def __init__(self, parent, object_cls):
-        self._parent = parent
+        self._parent_ref = weakref.ref(parent)
         self._object_cls = object_cls
         self._objects = OrderedDict()
 
+    @property
+    def _parent(self):
+        return self._parent_ref()
+
     def __setitem__(self, name, obj):
         self._objects[name] = obj
 
@@ -218,12 +245,16 @@ class _LazyObjectLookup(Mapping):
 
     def __iter__(self):
         for name in self._objects:
-            yield name
+            # fix variable name for variable which clashes with dim name
+            yield name.replace("_nc4_non_coord_", "")
 
     def __len__(self):
         return len(self._objects)
 
     def __getitem__(self, key):
+        # check for _nc4_non_coord_ variable
+        if key not in self._objects and "_nc4_non_coord_" + key in self._objects:
+            key = "_nc4_non_coord_" + key
         if self._objects[key] is not None:
             return self._objects[key]
         else:
@@ -256,8 +287,8 @@ class Group(Mapping):
         return Group
 
     def __init__(self, parent, name):
-        self._parent = parent
-        self._root = parent._root
+        self._parent_ref = weakref.ref(parent)
+        self._root_ref = weakref.ref(parent._root)
         self._h5path = _join_h5paths(parent.name, name)
 
         if parent is not self:
@@ -283,11 +314,13 @@ class Group(Mapping):
             else:
                 if v.attrs.get("CLASS") == b"DIMENSION_SCALE":
                     dim_id = v.attrs.get("_Netcdf4Dimid")
-                    if "_Netcdf4Coordinates" in v.attrs:
+                    # retrieve for _Netcdf4Coordinates with shape > 1
+                    if "_Netcdf4Coordinates" in v.attrs and len(v.shape) > 1:
                         assert dim_id is not None
                         coord_ids = v.attrs["_Netcdf4Coordinates"]
                         size = v.shape[list(coord_ids).index(dim_id)]
                         current_size = size
+                    # retrieve for 1D scale/variable
                     else:
                         assert len(v.shape) == 1
                         # Unlimited dimensions are represented as None.
@@ -321,11 +354,8 @@ class Group(Mapping):
                                 phony_dims[dimsize] = max(phony_dims[dimsize], cnt)
 
                 if not _netcdf_dimension_but_not_variable(v):
-                    if isinstance(v, h5py.Dataset):
-                        var_name = k
-                        if k.startswith("_nc4_non_coord_"):
-                            var_name = k[len("_nc4_non_coord_") :]
-                        self._variables.add(var_name)
+                    if isinstance(v, h5_dataset_types):
+                        self._variables.add(k)
 
         # iterate over found phony dimensions and create them
         if self._root._phony_dims_mode is not None:
@@ -344,6 +374,14 @@ class Group(Mapping):
 
         self._initialized = True
 
+    @property
+    def _root(self):
+        return self._root_ref()
+
+    @property
+    def _parent(self):
+        return self._parent_ref()
+
     def _create_phony_dimensions(self):
         # this is for 'sort' naming
         for key, value in self._phony_dims.items():
@@ -388,6 +426,19 @@ class Group(Mapping):
         # subclasses:
         return self._root._h5file[self._h5path]
 
+    @property
+    def _track_order(self):
+        # TODO: make a suggestion to upstream to create a property
+        # for files to get if they track the order
+        # As of version 3.6.0 this property did not exist
+        from h5py.h5p import CRT_ORDER_INDEXED, CRT_ORDER_TRACKED
+
+        gcpl = self._h5group.id.get_create_plist()
+        attr_creation_order = gcpl.get_attr_creation_order()
+        order_tracked = bool(attr_creation_order & CRT_ORDER_TRACKED)
+        order_indexed = bool(attr_creation_order & CRT_ORDER_INDEXED)
+        return order_tracked and order_indexed
+
     @property
     def name(self):
         return self._h5group.name
@@ -396,9 +447,14 @@ class Group(Mapping):
         if name in self._dim_sizes.maps[0]:
             raise ValueError("dimension %r already exists" % name)
 
-        self._dim_sizes[name] = size
+        self._dim_sizes[name] = None if size == 0 else size
         self._current_dim_sizes[name] = 0 if size is None else size
-        self._dim_order[name] = None
+        # Increase maximum dimension id (_Netcdf4Dimid)
+        self._root._max_dim_id += 1
+        self._dim_order[name] = self._root._max_dim_id
+        # Create dimension scale only if we have write permission
+        if self._root._writable and name not in self._h5group:
+            self._create_dim_scale(name)
 
     @property
     def dimensions(self):
@@ -419,7 +475,7 @@ class Group(Mapping):
     def _create_child_group(self, name):
         if name in self:
             raise ValueError("unable to create group %r (name already exists)" % name)
-        self._h5group.create_group(name)
+        self._h5group.create_group(name, track_order=self._track_order)
         self._groups[name] = self._group_cls(self, name)
         return self._groups[name]
 
@@ -441,8 +497,6 @@ class Group(Mapping):
     def _create_child_variable(
         self, name, dimensions, dtype, data, fillvalue, **kwargs
     ):
-        stacklevel = 4  # correct if name does not start with '/'
-
         if name in self:
             raise ValueError(
                 "unable to create variable %r " "(name already exists)" % name
@@ -461,22 +515,21 @@ class Group(Mapping):
             # never warn since h5netcdf has always errored here
             _invalid_netcdf_feature(
                 "boolean dtypes",
-                allow=bool(self._root.invalid_netcdf),
-                file=self._root,
-                stacklevel=stacklevel,
+                self._root.invalid_netcdf,
             )
         else:
-            self._root._check_valid_netcdf_dtype(dtype, stacklevel=stacklevel)
+            self._root._check_valid_netcdf_dtype(dtype)
 
         if "scaleoffset" in kwargs:
             _invalid_netcdf_feature(
                 "scale-offset filters",
-                allow=self._root.invalid_netcdf,
-                file=self._root,
-                stacklevel=stacklevel,
+                self._root.invalid_netcdf,
             )
 
-        if name in self.dimensions and name not in dimensions:
+        # variable <-> dimension name clash
+        if name in self.dimensions and (
+            name not in dimensions or (len(dimensions) > 1 and dimensions[0] != name)
+        ):
             h5name = "_nc4_non_coord_" + name
         else:
             h5name = name
@@ -491,19 +544,43 @@ class Group(Mapping):
 
         # Clear dummy HDF5 datasets with this name that were created for a
         # dimension scale without a corresponding variable.
-        if name in self.dimensions and name in self._h5group:
-            h5ds = self._h5group[name]
+        # Keep the references, to re-attach later
+        refs = None
+        if h5name in self.dimensions and h5name in self._h5group:
+            h5ds = self._h5group[h5name]
             if _netcdf_dimension_but_not_variable(h5ds):
-                self._detach_dim_scale(name)
-                del self._h5group[name]
+                refs = self._get_dim_scale_refs(name)
+                self._delete_dim_scale(name)
 
         self._h5group.create_dataset(
-            h5name, shape, dtype=dtype, data=data, fillvalue=fillvalue, **kwargs
+            h5name,
+            shape,
+            dtype=dtype,
+            data=data,
+            fillvalue=fillvalue,
+            track_order=self._track_order,
+            **kwargs,
         )
 
         self._variables[h5name] = self._variable_cls(self, h5name, dimensions)
         variable = self._variables[h5name]
 
+        # Re-create dim-scale and re-attach references to coordinate variable.
+        if name in self.dimensions and h5name in self._h5group:
+            self._create_dim_scale(name)
+            if refs is not None:
+                self._attach_dim_scale(name, refs)
+
+        # In case of data variables attach dim_scales and coords.
+        if name in self.variables and h5name not in self.dimensions:
+            variable._attach_dim_scales()
+            variable._attach_coords()
+
+        # This is a bit of a hack, netCDF4 attaches _Netcdf4Dimid to every variable
+        # when a variable is first written to, after variable creation.
+        # Here we just attach it to every variable on creation.
+        variable._ensure_dim_id()
+
         if fillvalue is not None:
             value = variable.dtype.type(fillvalue)
             variable.attrs._h5attrs["_FillValue"] = value
@@ -548,60 +625,72 @@ class Group(Mapping):
     def __len__(self):
         return len(self.variables) + len(self.groups)
 
-    def _create_dim_scales(self):
-        """Create all necessary HDF5 dimension scale."""
+    def _create_dim_scale(self, dim):
+        """Create HDF5 dimension scale."""
         dim_order = self._dim_order.maps[0]
-        for dim in sorted(dim_order, key=lambda d: dim_order[d]):
-            if dim not in self._h5group:
-                size = self._current_dim_sizes[dim]
-                kwargs = {}
-                if self._dim_sizes[dim] is None:
-                    kwargs["maxshape"] = (None,)
-                self._h5group.create_dataset(
-                    name=dim, shape=(size,), dtype=">f4", **kwargs
-                )
+        if dim not in self._h5group:
+            size = self._current_dim_sizes[dim]
+            kwargs = {}
+            if self._dim_sizes[dim] is None:
+                kwargs["maxshape"] = (None,)
+            self._h5group.create_dataset(
+                name=dim,
+                shape=(size,),
+                dtype=">f4",
+                track_order=self._track_order,
+                **kwargs,
+            )
 
-            h5ds = self._h5group[dim]
-            h5ds.attrs["_Netcdf4Dimid"] = np.int32(dim_order[dim])
+        h5ds = self._h5group[dim]
+        h5ds.attrs["_Netcdf4Dimid"] = np.array(dim_order[dim], dtype=np.int32)
 
-            if len(h5ds.shape) > 1:
-                dims = self._variables[dim].dimensions
-                coord_ids = np.array([dim_order[d] for d in dims], "int32")
-                h5ds.attrs["_Netcdf4Coordinates"] = coord_ids
+        if len(h5ds.shape) > 1:
+            dims = self._variables[dim].dimensions
+            coord_ids = np.array([dim_order[d] for d in dims], "int32")
+            h5ds.attrs["_Netcdf4Coordinates"] = coord_ids
 
-            # TODO: don't re-create scales if they already exist. With the
-            # current version of h5py, this would require using the low-level
-            # h5py.h5ds.is_scale interface to detect pre-existing scales.
-            dimlen = bytes(f"{self._current_dim_sizes[dim]:10}", "ascii")
-            scale_name = dim if dim in self.variables else NOT_A_VARIABLE + dimlen
-            if h5py.__version__ < LooseVersion("2.10.0"):
+        dimlen = bytes(f"{self._current_dim_sizes[dim]:10}", "ascii")
+        scale_name = dim if dim in self.variables else NOT_A_VARIABLE + dimlen
+        # don't re-create scales if they already exist.
+        if not h5py.h5ds.is_scale(h5ds.id):
+            if version.parse(h5py.__version__) < version.parse("2.10.0"):
                 h5ds.dims.create_scale(h5ds, scale_name)
             else:
                 h5ds.make_scale(scale_name)
 
-        for subgroup in self.groups.values():
-            subgroup._create_dim_scales()
+    def _attach_dim_scale(self, name, refs):
+        """Attach dimension scale to variable references"""
+        for var, dim in refs:
+            self._all_h5groups[var].dims[dim].attach_scale(self._all_h5groups[name])
 
-    def _attach_dim_scales(self):
-        """Attach dimension scales to all variables."""
-        for name, var in self.variables.items():
-            if name not in self.dimensions:
-                for n, dim in enumerate(var.dimensions):
-                    var._h5ds.dims[n].attach_scale(self._all_h5groups[dim])
+    def _detach_dim_scale(self, name, refs):
+        """Detach dimension scale from variable references"""
+        for var, dim in refs:
+            self._all_h5groups[var].dims[dim].detach_scale(self._all_h5groups[name])
 
-        for subgroup in self.groups.values():
-            subgroup._attach_dim_scales()
+    def _get_dim_scale_refs(self, name):
+        """Get variable scale references from dimension scale"""
+        return (
+            list(self._h5group[name].attrs.get("REFERENCE_LIST", []))
+            if name in self._h5group
+            else []
+        )
 
-    def _detach_dim_scale(self, name):
-        """Detach the dimension scale corresponding to a dimension name."""
-        for var in self.variables.values():
-            for n, dim in enumerate(var.dimensions):
-                if dim == name:
-                    var._h5ds.dims[n].detach_scale(self._all_h5groups[dim])
+    def _is_dimscale(self, name):
+        """Check if name is dimension scale"""
+        return (
+            _netcdf_dimension_but_not_variable(self._h5group[name])
+            if name in self._h5group
+            else False
+        )
 
-        for subgroup in self.groups.values():
-            if dim not in subgroup._h5group:
-                subgroup._detach_dim_scale(name)
+    def _delete_dim_scale(self, name):
+        """Delete Dimension Scale"""
+        if self._is_dimscale(name):
+            refs = self._get_dim_scale_refs(name)
+            if refs:
+                self._detach_dim_scale(name, refs)
+            del self._h5group[name]
 
     @property
     def parent(self):
@@ -656,7 +745,7 @@ class Group(Mapping):
         header = "<%s %r (%s members)>" % (self._cls_name, self.name, len(self))
         return "\n".join([header] + self._repr_body())
 
-    def resize_dimension(self, dimension, size):
+    def resize_dimension(self, dimension, size, resize_vars=True):
         """
         Resize a dimension to a certain size.
 
@@ -672,6 +761,17 @@ class Group(Mapping):
         # Resize the dimension.
         self._current_dim_sizes[dimension] = size
 
+        # Resize dimension in file
+        if dimension in self._h5group:
+            self._h5group[dimension].resize((size,))
+
+        # Adapt recursively to all variables
+        if resize_vars:
+            self._resize_variables(dimension, size)
+
+    def _resize_variables(self, dimension, size, recurse=True):
+        """Resize variables"""
+        # Adapt to all variables
         for var in self.variables.values():
             new_shape = list(var.shape)
             for i, d in enumerate(var.dimensions):
@@ -682,13 +782,81 @@ class Group(Mapping):
                 var._h5ds.resize(new_shape)
 
         # Recurse as dimensions are visible to this group and all child groups.
-        for i in self.groups.values():
-            i.resize_dimension(dimension, size)
+        if recurse:
+            for i in self.groups.values():
+                i._resize_variables(dimension, size)
 
 
 class File(Group):
-    def __init__(self, path, mode="a", invalid_netcdf=None, phony_dims=None, **kwargs):
-        if h5py.__version__ >= LooseVersion("3.0.0"):
+    def __init__(
+        self, path, mode=None, invalid_netcdf=False, phony_dims=None, **kwargs
+    ):
+        """NetCDF4 file constructor.
+
+        Parameters
+        ----------
+        path: path-like
+            Location of the netCDF4 file to be accessed.
+
+        mode: "r", "r+", "a", "w"
+            A valid file access mode. See
+
+        invalid_netcdf: bool
+            Allow writing netCDF4 with data types and attributes that would
+            otherwise not generate netCDF4 files that can be read by other
+            applications. See
+            https://github.com/h5netcdf/h5netcdf#invalid-netcdf-files
+            for more details.
+
+        phony_dims: 'sort', 'access'
+            See:
+            https://github.com/h5netcdf/h5netcdf#datasets-with-missing-dimension-scales
+
+        **kwargs:
+            Additional keyword arguments to be passed to the ``h5py.File``
+            constructor.
+
+        Notes
+        -----
+        In h5netcdf version 0.12.0 and earlier, order tracking was disabled in
+        HDF5 file. As this is a requirement for the current netCDF4 standard,
+        it has been enabled without deprecation as of version 0.13.0 [1]_.
+
+        Datasets created with h5netcdf version 0.12.0 that are opened with
+        newer versions of h5netcdf will continue to disable order tracker.
+
+        .. [1] https://github.com/h5netcdf/h5netcdf/issues/128
+
+        """
+        # 2022/01/09
+        # netCDF4 wants the track_order parameter to be true
+        # through this might be getting relaxed in a more recent version of the
+        # standard
+        # https://github.com/Unidata/netcdf-c/issues/2054
+        # https://github.com/h5netcdf/h5netcdf/issues/128
+        track_order = kwargs.pop("track_order", True)
+
+        if not track_order:
+            raise ValueError(
+                f"track_order, if specified must be set to to True (got {track_order})"
+                "to conform to the netCDF4 file format. Please see "
+                "https://github.com/h5netcdf/h5netcdf/issues/130 "
+                "for more details."
+            )
+
+        # Deprecating mode='a' in favor of mode='r'
+        # If mode is None default to 'a' and issue a warning
+        if mode is None:
+            msg = (
+                "Falling back to mode='a'. "
+                "In future versions, mode will default to read-only. "
+                "It is recommended to explicitly set mode='r' to prevent any unintended "
+                "changes to the file."
+            )
+            warnings.warn(msg, FutureWarning, stacklevel=2)
+            mode = "a"
+
+        if version.parse(h5py.__version__) >= version.parse("3.0.0"):
             self.decode_vlen_strings = kwargs.pop("decode_vlen_strings", None)
         try:
             if isinstance(path, str):
@@ -704,19 +872,25 @@ class File(Group):
                         self._preexisting_file = True
                     except IOError:
                         self._preexisting_file = False
-                    self._h5file = h5pyd.File(path, mode, **kwargs)
+                    self._h5file = h5pyd.File(
+                        path, mode, track_order=track_order, **kwargs
+                    )
                 else:
-                    self._preexisting_file = os.path.exists(path)
-                    self._h5file = h5py.File(path, mode, **kwargs)
+                    self._preexisting_file = os.path.exists(path) and mode != "w"
+                    self._h5file = h5py.File(
+                        path, mode, track_order=track_order, **kwargs
+                    )
             else:  # file-like object
-                if h5py.__version__ < LooseVersion("2.9.0"):
+                if version.parse(h5py.__version__) < version.parse("2.9.0"):
                     raise TypeError(
                         "h5py version ({}) must be greater than 2.9.0 to load "
                         "file-like objects.".format(h5py.__version__)
                     )
                 else:
                     self._preexisting_file = mode in {"r", "r+", "a"}
-                    self._h5file = h5py.File(path, mode, **kwargs)
+                    self._h5file = h5py.File(
+                        path, mode, track_order=track_order, **kwargs
+                    )
         except Exception:
             self._closed = True
             raise
@@ -724,12 +898,10 @@ class File(Group):
             self._closed = False
 
         self._mode = mode
-        self._root = self
+        self._writable = mode != "r"
+        self._root_ref = weakref.ref(self)
         self._h5path = "/"
         self.invalid_netcdf = invalid_netcdf
-        # If invalid_netcdf is None, we'll disable writing _NCProperties only
-        # if we actually use invalid NetCDF features.
-        self._write_ncproperties = invalid_netcdf is not True
 
         # phony dimension handling
         self._phony_dims_mode = phony_dims
@@ -744,7 +916,7 @@ class File(Group):
                 )
 
         # string decoding
-        if h5py.__version__ >= LooseVersion("3.0.0"):
+        if version.parse(h5py.__version__) >= version.parse("3.0.0"):
             if "legacy" in self._cls_name:
                 if self.decode_vlen_strings is not None:
                     msg = (
@@ -757,18 +929,24 @@ class File(Group):
                 if self.decode_vlen_strings is None:
                     msg = (
                         "String decoding changed with h5py >= 3.0. "
+                        "See https://docs.h5py.org/en/latest/strings.html and "
+                        "https://github.com/h5netcdf/h5netcdf/issues/132 for more details. "
                         "Currently backwards compatibility with h5py < 3.0 is kept by "
                         "decoding vlen strings per default. This will change in future "
-                        "versions for consistency with h5py >= 3.0. "
-                        "Setting 'decode_vlen_strings=True' forces string decoding."
+                        "versions for consistency with h5py >= 3.0. To silence this "
+                        "warning set kwarg ``decode_vlen_strings=False`` which will "
+                        "return Python bytes from variables containing vlen strings. Setting "
+                        "``decode_vlen_strings=True`` forces vlen string decoding which returns "
+                        "Python strings from variables containing vlen strings."
                     )
-                    warnings.warn(msg, FutureWarning, stacklevel=0)
+                    warnings.warn(msg, FutureWarning, stacklevel=2)
                     self.decode_vlen_strings = True
 
         # These maps keep track of dimensions in terms of size (might be
         # unlimited), current size (identical to size for limited dimensions),
         # their position, and look-up for HDF5 datasets corresponding to a
         # dimension.
+        self._max_dim_id = -1
         self._dim_sizes = ChainMap()
         self._current_dim_sizes = ChainMap()
         self._dim_order = ChainMap()
@@ -778,6 +956,18 @@ class File(Group):
         # mimics netcdf-c style naming
         if phony_dims == "sort":
             self._determine_phony_dimensions()
+        # get maximum dimension id
+        if self._writable:
+            self._max_dim_id = self._get_maximum_dimension_id()
+
+    def _get_maximum_dimension_id(self):
+        dimids = []
+
+        def _dimids(name, obj):
+            dimids.append(obj.attrs.get("_Netcdf4Dimid", -1))
+
+        self._h5file.visititems(_dimids)
+        return max(dimids) if dimids else -1
 
     def _determine_phony_dimensions(self):
         def get_labeled_dimension_count(grp):
@@ -794,7 +984,7 @@ class File(Group):
         self._labeled_dim_count = get_labeled_dimension_count(self)
         create_phony_dimensions(self)
 
-    def _check_valid_netcdf_dtype(self, dtype, stacklevel=3):
+    def _check_valid_netcdf_dtype(self, dtype):
         dtype = np.dtype(dtype)
 
         if dtype == bool:
@@ -813,9 +1003,7 @@ class File(Group):
         if description is not None:
             _invalid_netcdf_feature(
                 "{} dtypes".format(description),
-                allow=self.invalid_netcdf,
-                file=self,
-                stacklevel=stacklevel + 1,
+                self.invalid_netcdf,
             )
 
     @property
@@ -830,36 +1018,9 @@ class File(Group):
     def parent(self):
         return None
 
-    def _set_unassigned_dimension_ids(self):
-        max_dim_id = -1
-
-        # collect the largest assigned dimension ID
-        groups = [self]
-        while groups:
-            group = groups.pop()
-            assigned_dim_ids = [
-                dim_id for dim_id in group._dim_order.values() if dim_id is not None
-            ]
-            max_dim_id = max([max_dim_id] + assigned_dim_ids)
-            groups.extend(group._groups.values())
-
-        # set all dimension IDs to valid values
-        next_dim_id = max_dim_id + 1
-        groups = [self]
-        while groups:
-            group = groups.pop()
-            for key in group._dim_order:
-                if group._dim_order[key] is None:
-                    group._dim_order[key] = next_dim_id
-                    next_dim_id += 1
-            groups.extend(group._groups.values())
-
     def flush(self):
-        if "r" not in self._mode:
-            self._set_unassigned_dimension_ids()
-            self._create_dim_scales()
-            self._attach_dim_scales()
-            if not self._preexisting_file and self._write_ncproperties:
+        if self._writable:
+            if not self._preexisting_file and not self.invalid_netcdf:
                 self.attrs._h5attrs["_NCProperties"] = np.array(
                     _NC_PROPERTIES,
                     dtype=h5py.string_dtype(
