@@ -4,13 +4,13 @@ import random
 import re
 import string
 import tempfile
-from distutils.version import LooseVersion
 from os import environ as env
 
 import h5py
 import netCDF4
 import numpy as np
 import pytest
+from packaging import version
 from pytest import raises
 
 import h5netcdf
@@ -62,7 +62,7 @@ def tmp_local_or_remote_netcdf(request, tmpdir, restapi):
 
 @pytest.fixture(params=[True, False])
 def decode_vlen_strings(request):
-    if h5py.__version__ >= LooseVersion("3.0.0"):
+    if version.parse(h5py.__version__) >= version.parse("3.0.0"):
         return dict(decode_vlen_strings=request.param)
     else:
         return {}
@@ -450,7 +450,7 @@ def test_write_legacyapi_read_h5netcdf(tmp_local_netcdf, decode_vlen_strings):
 
 
 def test_fileobj(decode_vlen_strings):
-    if h5py.__version__ < LooseVersion("2.9.0"):
+    if version.parse(h5py.__version__) < version.parse("2.9.0"):
         pytest.skip("h5py > 2.9.0 required to test file-like objects")
     fileobj = tempfile.TemporaryFile()
     write_h5netcdf(fileobj)
@@ -498,16 +498,17 @@ def test_repr(tmp_local_or_remote_netcdf):
 
 
 def test_attrs_api(tmp_local_or_remote_netcdf):
-    with h5netcdf.File(tmp_local_or_remote_netcdf) as ds:
+    with h5netcdf.File(tmp_local_or_remote_netcdf, "w") as ds:
         ds.attrs["conventions"] = "CF"
+        ds.attrs["empty_string"] = h5py.Empty(dtype=np.dtype("|S1"))
         ds.dimensions["x"] = 1
         v = ds.create_variable("x", ("x",), "i4")
         v.attrs.update({"units": "meters", "foo": "bar"})
     assert ds._closed
-    with h5netcdf.File(tmp_local_or_remote_netcdf) as ds:
-        assert len(ds.attrs) == 1
-        assert dict(ds.attrs) == {"conventions": "CF"}
-        assert list(ds.attrs) == ["conventions"]
+    with h5netcdf.File(tmp_local_or_remote_netcdf, "r") as ds:
+        assert len(ds.attrs) == 2
+        assert dict(ds.attrs) == {"conventions": "CF", "empty_string": b""}
+        assert list(ds.attrs) == ["conventions", "empty_string"]
         assert dict(ds["x"].attrs) == {"units": "meters", "foo": "bar"}
         assert len(ds["x"].attrs) == 2
         assert sorted(ds["x"].attrs) == ["foo", "units"]
@@ -520,7 +521,7 @@ def test_optional_netcdf4_attrs(tmp_local_or_remote_netcdf):
         f.create_dataset("foo", data=foo_data)
         f.create_dataset("x", data=np.arange(5))
         f.create_dataset("y", data=np.arange(10))
-        if h5py.__version__ < LooseVersion("2.10.0"):
+        if version.parse(h5py.__version__) < version.parse("2.10.0"):
             f["foo"].dims.create_scale(f["x"])
             f["foo"].dims.create_scale(f["y"])
         else:
@@ -552,7 +553,8 @@ def test_error_handling(tmp_local_or_remote_netcdf):
 
 
 @pytest.mark.skipif(
-    h5py.__version__ < LooseVersion("3.0.0"), reason="not needed with h5py < 3.0"
+    version.parse(h5py.__version__) < version.parse("3.0.0"),
+    reason="not needed with h5py < 3.0",
 )
 def test_decode_string_warning(tmp_local_or_remote_netcdf):
     write_h5netcdf(tmp_local_or_remote_netcdf)
@@ -562,7 +564,8 @@ def test_decode_string_warning(tmp_local_or_remote_netcdf):
 
 
 @pytest.mark.skipif(
-    h5py.__version__ < LooseVersion("3.0.0"), reason="not needed with h5py < 3.0"
+    version.parse(h5py.__version__) < version.parse("3.0.0"),
+    reason="not needed with h5py < 3.0",
 )
 def test_decode_string_error(tmp_local_or_remote_netcdf):
     write_h5netcdf(tmp_local_or_remote_netcdf)
@@ -573,6 +576,12 @@ def test_decode_string_error(tmp_local_or_remote_netcdf):
             assert ds.name == "/"
 
 
+def test_mode_warning(tmp_local_or_remote_netcdf):
+    with pytest.warns(FutureWarning):
+        with h5netcdf.File(tmp_local_or_remote_netcdf):
+            pass
+
+
 def create_invalid_netcdf_data():
     foo_data = np.arange(125).reshape(5, 5, 5)
     bar_data = np.arange(625).reshape(25, 5, 5)
@@ -628,7 +637,7 @@ def test_invalid_netcdf4(tmp_local_or_remote_netcdf):
             var = dsr[grp].variables
             check_invalid_netcdf4(var, i)
 
-    with netCDF4.Dataset(tmp_local_or_remote_netcdf, mode="r") as dsr:
+    with netCDF4.Dataset(tmp_local_or_remote_netcdf, "r") as dsr:
         for i, grp in enumerate(grps):
             print(dsr[grp])
             var = dsr[grp].variables
@@ -674,7 +683,7 @@ def test_invalid_netcdf4_mixed(tmp_local_or_remote_netcdf):
         for k, v in var2.items():
             f.create_dataset(k, data=np.arange(v))
 
-        if h5py.__version__ < LooseVersion("2.10.0"):
+        if version.parse(h5py.__version__) < version.parse("2.10.0"):
             f["foo2"].dims.create_scale(f["x1"])
             f["foo2"].dims.create_scale(f["y1"])
             f["foo2"].dims.create_scale(f["z1"])
@@ -712,7 +721,7 @@ def test_invalid_netcdf_malformed_dimension_scales(tmp_local_or_remote_netcdf):
         f.create_dataset("y", data=np.arange(5))
         f.create_dataset("z", data=np.arange(5))
 
-        if h5py.__version__ < LooseVersion("2.10.0"):
+        if version.parse(h5py.__version__) < version.parse("2.10.0"):
             f["foo1"].dims.create_scale(f["x"])
             f["foo1"].dims.create_scale(f["y"])
             f["foo1"].dims.create_scale(f["z"])
@@ -776,7 +785,7 @@ def test_reading_str_array_from_netCDF4(tmp_local_netcdf, decode_vlen_strings):
 
 
 def test_nc_properties_new(tmp_local_or_remote_netcdf):
-    with h5netcdf.File(tmp_local_or_remote_netcdf):
+    with h5netcdf.File(tmp_local_or_remote_netcdf, "w"):
         pass
     h5 = get_hdf5_module(tmp_local_or_remote_netcdf)
     with h5.File(tmp_local_or_remote_netcdf, "r") as f:
@@ -829,28 +838,6 @@ def test_create_variable_matching_saved_dimension(tmp_local_or_remote_netcdf):
         assert f["y"].dims[0].keys() == ["x"]
 
 
-def test_invalid_netcdf_warns(tmp_local_or_remote_netcdf):
-    if tmp_local_or_remote_netcdf.startswith(remote_h5):
-        pytest.skip("h5pyd does not support NumPy complex dtype yet")
-    with h5netcdf.File(tmp_local_or_remote_netcdf, "w") as f:
-        # valid
-        with pytest.warns(None) as record:
-            f.create_variable(
-                "lzf_compressed", data=[1], dimensions=("x"), compression="lzf"
-            )
-        assert not record.list
-        # invalid
-        with pytest.warns(FutureWarning):
-            f.create_variable("complex", data=1j)
-        with pytest.warns(FutureWarning):
-            f.attrs["complex_attr"] = 1j
-        with pytest.warns(FutureWarning):
-            f.create_variable("scaleoffset", data=[1], dimensions=("x",), scaleoffset=0)
-    h5 = get_hdf5_module(tmp_local_or_remote_netcdf)
-    with h5.File(tmp_local_or_remote_netcdf, "r") as f:
-        assert "_NCProperties" not in f.attrs
-
-
 def test_invalid_netcdf_error(tmp_local_or_remote_netcdf):
     with h5netcdf.File(tmp_local_or_remote_netcdf, "w", invalid_netcdf=False) as f:
         # valid
@@ -888,18 +875,18 @@ def test_invalid_netcdf_okay(tmp_local_or_remote_netcdf):
 
 def test_reopen_file_different_dimension_sizes(tmp_local_netcdf):
     # regression test for https://github.com/h5netcdf/h5netcdf/issues/55
-    with h5netcdf.File(tmp_local_netcdf, mode="w") as f:
+    with h5netcdf.File(tmp_local_netcdf, "w") as f:
         f.create_variable("/one/foo", data=[1], dimensions=("x",))
-    with h5netcdf.File(tmp_local_netcdf, mode="a") as f:
+    with h5netcdf.File(tmp_local_netcdf, "a") as f:
         f.create_variable("/two/foo", data=[1, 2], dimensions=("x",))
-    with netCDF4.Dataset(tmp_local_netcdf, mode="r") as f:
+    with netCDF4.Dataset(tmp_local_netcdf, "r") as f:
         assert f.groups["one"].variables["foo"][...].shape == (1,)
 
 
 def test_invalid_then_valid_no_ncproperties(tmp_local_or_remote_netcdf):
     with h5netcdf.File(tmp_local_or_remote_netcdf, "w", invalid_netcdf=True):
         pass
-    with h5netcdf.File(tmp_local_or_remote_netcdf, "r"):
+    with h5netcdf.File(tmp_local_or_remote_netcdf, "a"):
         pass
     h5 = get_hdf5_module(tmp_local_or_remote_netcdf)
     with h5.File(tmp_local_or_remote_netcdf, "r") as f:
@@ -932,7 +919,7 @@ def test_creating_and_resizing_unlimited_dimensions(tmp_local_or_remote_netcdf):
 
 
 def test_creating_variables_with_unlimited_dimensions(tmp_local_or_remote_netcdf):
-    with h5netcdf.File(tmp_local_or_remote_netcdf) as f:
+    with h5netcdf.File(tmp_local_or_remote_netcdf, "w") as f:
         f.dimensions["x"] = None
         f.dimensions["y"] = 2
 
@@ -965,7 +952,7 @@ def test_creating_variables_with_unlimited_dimensions(tmp_local_or_remote_netcdf
 
     # Close and read again to also test correct parsing of unlimited
     # dimensions.
-    with h5netcdf.File(tmp_local_or_remote_netcdf) as f:
+    with h5netcdf.File(tmp_local_or_remote_netcdf, "r") as f:
         assert f.dimensions["x"] is None
         assert f._h5file["x"].maxshape == (None,)
         assert f._h5file["x"].shape == (3,)
@@ -976,7 +963,7 @@ def test_creating_variables_with_unlimited_dimensions(tmp_local_or_remote_netcdf
 
 
 def test_writing_to_an_unlimited_dimension(tmp_local_or_remote_netcdf):
-    with h5netcdf.File(tmp_local_or_remote_netcdf) as f:
+    with h5netcdf.File(tmp_local_or_remote_netcdf, "w") as f:
         # Two dimensions, only one is unlimited.
         f.dimensions["x"] = None
         f.dimensions["y"] = 3
@@ -1016,7 +1003,7 @@ def test_writing_to_an_unlimited_dimension(tmp_local_or_remote_netcdf):
 
 
 def test_c_api_can_read_unlimited_dimensions(tmp_local_netcdf):
-    with h5netcdf.File(tmp_local_netcdf) as f:
+    with h5netcdf.File(tmp_local_netcdf, "w") as f:
         # Three dimensions, only one is limited.
         f.dimensions["x"] = None
         f.dimensions["y"] = 3
@@ -1028,7 +1015,7 @@ def test_c_api_can_read_unlimited_dimensions(tmp_local_netcdf):
         g.create_variable("dummy4", dimensions=("z", "z"), dtype=np.int64)
         f.resize_dimension("x", 2)
 
-    with netCDF4.Dataset(tmp_local_netcdf) as f:
+    with netCDF4.Dataset(tmp_local_netcdf, "r") as f:
         assert f.dimensions["x"].size == 2
         assert f.dimensions["x"].isunlimited() is True
         assert f.dimensions["y"].size == 3
@@ -1058,7 +1045,7 @@ def test_reading_unlimited_dimensions_created_with_c_api(tmp_local_netcdf):
         # Assign something to trigger a resize.
         dummy1[:] = [[1, 2, 3], [4, 5, 6]]
 
-    with h5netcdf.File(tmp_local_netcdf) as f:
+    with h5netcdf.File(tmp_local_netcdf, "r") as f:
         assert f.dimensions["x"] is None
         assert f.dimensions["y"] == 3
         assert f.dimensions["z"] is None
@@ -1084,13 +1071,371 @@ def test_reading_unused_unlimited_dimension(tmp_local_or_remote_netcdf):
         f.resize_dimension("x", 5)
         assert f.dimensions == {"x": None}
 
-    f = h5netcdf.File(tmp_local_or_remote_netcdf, "r")
-
 
 def test_reading_special_datatype_created_with_c_api(tmp_local_netcdf):
     """Test reading a file with unsupported Datatype"""
     with netCDF4.Dataset(tmp_local_netcdf, "w") as f:
         complex128 = np.dtype([("real", np.float64), ("imag", np.float64)])
         f.createCompoundType(complex128, "complex128")
-    with h5netcdf.File(tmp_local_netcdf) as f:
+    with h5netcdf.File(tmp_local_netcdf, "r") as f:
         pass
+
+
+def test_nc4_non_coord(tmp_local_netcdf):
+    # Track order True is the new default for versions after 0.12.0
+    # 0.12.0 defaults to `track_order=False`
+    # Ensure that the tests order the variables in their creation order
+    # not alphabetical order
+    with h5netcdf.File(tmp_local_netcdf, "w") as f:
+        f.dimensions = {"x": None, "y": 2}
+        f.create_variable("test", dimensions=("x",), dtype=np.int64)
+        f.create_variable("y", dimensions=("x",), dtype=np.int64)
+
+    with h5netcdf.File(tmp_local_netcdf, "r") as f:
+        assert f.dimensions == {"x": None, "y": 2}
+        assert list(f.variables) == ["test", "y"]
+        assert list(f._h5group.keys()) == ["x", "y", "test", "_nc4_non_coord_y"]
+
+
+def test_overwrite_existing_file(tmp_local_netcdf):
+    # create file with _NCProperties attribute
+    with netCDF4.Dataset(tmp_local_netcdf, "w") as ds:
+        ds.createDimension("x", 10)
+
+    # check attribute
+    with h5netcdf.File(tmp_local_netcdf, "r") as ds:
+        assert ds.attrs._h5attrs.get("_NCProperties", False)
+
+    # overwrite file with legacyapi
+    with legacyapi.Dataset(tmp_local_netcdf, "w") as ds:
+        ds.createDimension("x", 10)
+
+    # check attribute
+    with h5netcdf.File(tmp_local_netcdf, "r") as ds:
+        assert ds.attrs._h5attrs.get("_NCProperties", False)
+
+    # overwrite file with new api
+    with h5netcdf.File(tmp_local_netcdf, "w") as ds:
+        ds.dimensions["x"] = 10
+
+    # check attribute
+    with h5netcdf.File(tmp_local_netcdf, "r") as ds:
+        assert ds.attrs._h5attrs.get("_NCProperties", False)
+
+
+def test_scales_on_append(tmp_local_netcdf):
+    # create file with _NCProperties attribute
+    with netCDF4.Dataset(tmp_local_netcdf, "w") as ds:
+        ds.createDimension("x", 10)
+
+    # append file with netCDF4
+    with netCDF4.Dataset(tmp_local_netcdf, "r+") as ds:
+        ds.createVariable("test", "i4", ("x",))
+
+    # check scales
+    with h5netcdf.File(tmp_local_netcdf, "r") as ds:
+        assert ds.variables["test"].attrs._h5attrs.get("DIMENSION_LIST", False)
+
+    # append file with legacyapi
+    with legacyapi.Dataset(tmp_local_netcdf, "r+") as ds:
+        ds.createVariable("test1", "i4", ("x",))
+
+    # check scales
+    with h5netcdf.File(tmp_local_netcdf, "r") as ds:
+        assert ds.variables["test1"].attrs._h5attrs.get("DIMENSION_LIST", False)
+
+
+def create_attach_scales(filename, append_module):
+    # create file with netCDF4
+    with netCDF4.Dataset(filename, "w") as ds:
+        ds.createDimension("x", 0)
+        ds.createDimension("y", 1)
+        ds.createVariable("test", "i4", ("x",))
+        ds.variables["test"] = np.ones((10,))
+
+    # append file with netCDF4
+    with append_module.Dataset(filename, "a") as ds:
+        ds.createVariable("test1", "i4", ("x",))
+        ds.createVariable("y", "i4", ("x", "y"))
+
+    # check scales
+    with h5netcdf.File(filename, "r") as ds:
+        refs = ds._h5group["x"].attrs.get("REFERENCE_LIST", False)
+        assert len(refs) == 3
+        for (ref, dim), name in zip(refs, ["/test", "/test1", "/_nc4_non_coord_y"]):
+            assert dim == 0
+            assert ds._root._h5file[ref].name == name
+
+
+def test_create_attach_scales_netcdf4(tmp_local_netcdf):
+    create_attach_scales(tmp_local_netcdf, netCDF4)
+
+
+def test_create_attach_scales_legacyapi(tmp_local_netcdf):
+    create_attach_scales(tmp_local_netcdf, legacyapi)
+
+
+def test_detach_scale(tmp_local_netcdf):
+    with h5netcdf.File(tmp_local_netcdf, "w") as ds:
+        ds.dimensions["x"] = 2
+        ds.dimensions["y"] = 2
+
+    with h5netcdf.File(tmp_local_netcdf, "a") as ds:
+        ds.create_variable("test", dimensions=("x",), dtype=np.int64)
+        # this forces detach and re-creation
+        ds.create_variable("x", dimensions=("y",), dtype=np.int64)
+
+    with h5netcdf.File(tmp_local_netcdf, "r") as ds:
+        refs = ds._h5group["x"].attrs.get("REFERENCE_LIST", False)
+        assert len(refs) == 1
+        for (ref, dim), name in zip(refs, ["/test"]):
+            assert dim == 0
+            assert ds._root._h5file[ref].name == name
+
+
+def test_is_dimscale(tmp_local_netcdf):
+    with legacyapi.Dataset(tmp_local_netcdf, "w") as ds:
+        ds.createDimension("x", 10)
+    with legacyapi.Dataset(tmp_local_netcdf, "r") as ds:
+        assert ds._is_dimscale("x")
+
+
+def test_delete_dim_scale(tmp_local_netcdf):
+    with legacyapi.Dataset(tmp_local_netcdf, "w") as ds:
+        ds.createDimension("x", 10)
+    with legacyapi.Dataset(tmp_local_netcdf, "r+") as ds:
+        assert ds._is_dimscale("x")
+        ds._delete_dim_scale("x")
+        assert not ds._is_dimscale("x")
+
+
+def test_get_dim_scale_refs(tmp_local_netcdf):
+    with legacyapi.Dataset(tmp_local_netcdf, "w") as ds:
+        ds.createDimension("x", 10)
+        ds.createVariable("test0", "i8", ("x",))
+        ds.createVariable("test1", "i8", ("x",))
+    with legacyapi.Dataset(tmp_local_netcdf, "r") as ds:
+        refs = ds._get_dim_scale_refs("x")
+        assert ds._h5file[refs[0][0]] == ds["test0"]._h5ds
+        assert ds._h5file[refs[1][0]] == ds["test1"]._h5ds
+
+
+def create_netcdf_dimensions(ds, idx):
+    # dimension and variable setup is adapted from the blogpost at
+    # https://www.unidata.ucar.edu/blogs/developer/en/entry/netcdf4_shared_dimensions
+    g = ds.createGroup("dimtest" + str(idx))
+    g.createDimension("time", 0)  # time
+    g.createDimension("nvec", 5 + idx)  # nvec
+    g.createDimension("sample", 2 + idx)  # sample
+    g.createDimension("ship", 3 + idx)  # ship
+    g.createDimension("ship_strlen", 10)  # ship_strlen
+    g.createDimension("collide", 7 + idx)  # collide
+
+    time = g.createVariable("time", "f8", ("time",))
+    data = g.createVariable("data", "i8", ("ship", "sample", "time", "nvec"))
+    collide = g.createVariable("collide", "i8", ("nvec",))
+    non_collide = g.createVariable("non_collide", "i8", ("nvec",))
+    ship = g.createVariable("ship", "S1", ("ship", "ship_strlen"))
+    sample = g.createVariable("sample", "i8", ("time", "sample"))
+
+    if isinstance(ds, legacyapi.Dataset):
+        g.resize_dimension("time", 10 + idx)
+    time[:] = np.arange(10 + idx)
+    data[:] = np.ones((3 + idx, 2 + idx, 10 + idx, 5 + idx)) * 12.0
+    collide[...] = np.arange(5 + idx)
+    non_collide[...] = np.arange(5 + idx) + 10
+    sample[0 : 2 + idx, : 2 + idx] = np.ones((2 + idx, 2 + idx))
+    if version.parse(h5py.__version__) >= version.parse("3.0.0"):
+        ship[0] = list("Skiff     ")
+    else:
+        ship[0] = string_to_char(np.array("Skiff     ", dtype="|S1"))
+
+
+def create_h5netcdf_dimensions(ds, idx):
+    # dimension and variable setup is adapted from the blogpost at
+    # https://www.unidata.ucar.edu/blogs/developer/en/entry/netcdf4_shared_dimensions
+    g = ds.create_group("dimtest" + str(idx))
+    g.dimensions["time"] = 0  # time
+    g.dimensions["nvec"] = 5 + idx  # nvec
+    g.dimensions["sample"] = 2 + idx  # sample
+    g.dimensions["ship"] = 3 + idx  # ship
+    g.dimensions["ship_strlen"] = 10  # ship_strlen
+    g.dimensions["collide"] = 7 + idx  # collide
+
+    g.create_variable("time", dimensions=("time",), dtype=np.float64)
+    g.create_variable(
+        "data", dimensions=("ship", "sample", "time", "nvec"), dtype=np.int64
+    )
+    g.create_variable("collide", dimensions=("nvec",), dtype=np.int64)
+    g.create_variable("non_collide", dimensions=("nvec",), dtype=np.int64)
+    g.create_variable("sample", dimensions=("time", "sample"), dtype=np.int64)
+    g.create_variable("ship", dimensions=("ship", "ship_strlen"), dtype="S1")
+
+    g.resize_dimension("time", 10 + idx)
+    g.variables["time"][:] = np.arange(10 + idx)
+    g.variables["data"][:] = np.ones((3 + idx, 2 + idx, 10 + idx, 5 + idx)) * 12.0
+    g.variables["collide"][...] = np.arange(5 + idx)
+    g.variables["non_collide"][...] = np.arange(5 + idx) + 10
+    g.variables["sample"][0 : 2 + idx, : 2 + idx] = np.ones((2 + idx, 2 + idx))
+    if version.parse(h5py.__version__) >= version.parse("3.0.0"):
+        g.variables["ship"][0] = list("Skiff     ")
+    else:
+        g.variables["ship"][0] = string_to_char(np.array("Skiff     ", dtype="|S1"))
+
+
+def check_netcdf_dimensions(tmp_netcdf, write_module, read_module):
+    if read_module in [legacyapi, netCDF4]:
+        opener = read_module.Dataset
+    else:
+        opener = h5netcdf.File
+    with opener(tmp_netcdf, "r") as ds:
+        for i, grp in enumerate(["dimtest0", "dimtest1"]):
+            g = ds.groups[grp]
+            assert set(g.dimensions) == {
+                "collide",
+                "ship_strlen",
+                "time",
+                "nvec",
+                "ship",
+                "sample",
+            }
+            if read_module in [legacyapi, h5netcdf]:
+                assert g.dimensions["time"] is None
+                assert g._current_dim_sizes["time"] == 10 + i
+                assert g.dimensions["nvec"] is not None
+                assert g.dimensions["nvec"] == 5 + i
+                assert g.dimensions["sample"] is not None
+                assert g.dimensions["sample"] == 2 + i
+                assert g.dimensions["collide"] is not None
+                assert g.dimensions["collide"] == 7 + i
+                assert g.dimensions["ship"] is not None
+                assert g.dimensions["ship"] == 3 + i
+                assert g.dimensions["ship_strlen"] is not None
+                assert g.dimensions["ship_strlen"] == 10
+            else:
+                assert g.dimensions["time"].isunlimited()
+                assert g.dimensions["time"].size == 10 + i
+                assert not g.dimensions["nvec"].isunlimited()
+                assert g.dimensions["nvec"].size == 5 + i
+                assert not g.dimensions["sample"].isunlimited()
+                assert g.dimensions["sample"].size == 2 + i
+                assert not g.dimensions["ship"].isunlimited()
+                assert g.dimensions["ship"].size == 3 + i
+                assert not g.dimensions["ship_strlen"].isunlimited()
+                assert g.dimensions["ship_strlen"].size == 10
+                assert not g.dimensions["collide"].isunlimited()
+                assert g.dimensions["collide"].size == 7 + i
+
+            assert set(g.variables) == {
+                "data",
+                "collide",
+                "non_collide",
+                "time",
+                "sample",
+                "ship",
+            }
+            assert g.variables["time"].shape == (10 + i,)
+            assert g.variables["data"].shape == (3 + i, 2 + i, 10 + i, 5 + i)
+            assert g.variables["collide"].shape == (5 + i,)
+            assert g.variables["non_collide"].shape == (5 + i,)
+            if write_module == netCDF4 and read_module in [legacyapi, h5netcdf]:
+                assert g.variables["sample"].shape == (2 + i, 2 + i)
+            else:
+                assert g.variables["sample"].shape == (10 + i, 2 + i)
+            assert g.variables["ship"].shape == (3 + i, 10)
+
+
+def write_dimensions(tmp_netcdf, write_module):
+    if write_module in [legacyapi, netCDF4]:
+        with write_module.Dataset(tmp_netcdf, "w") as ds:
+            create_netcdf_dimensions(ds, 0)
+            create_netcdf_dimensions(ds, 1)
+    else:
+        with write_module.File(tmp_netcdf, "w") as ds:
+            create_h5netcdf_dimensions(ds, 0)
+            create_h5netcdf_dimensions(ds, 1)
+
+
+@pytest.fixture(
+    params=[
+        [netCDF4, netCDF4],
+        [legacyapi, legacyapi],
+        [h5netcdf, h5netcdf],
+        [legacyapi, netCDF4],
+        [netCDF4, legacyapi],
+        [h5netcdf, netCDF4],
+        [netCDF4, h5netcdf],
+        [legacyapi, h5netcdf],
+        [h5netcdf, legacyapi],
+    ]
+)
+def read_write_matrix(request):
+    print("write module:", request.param[0].__name__)
+    print("read_module:", request.param[1].__name__)
+    return request.param
+
+
+def test_dimensions(tmp_local_netcdf, read_write_matrix):
+    write_dimensions(tmp_local_netcdf, read_write_matrix[0])
+    check_netcdf_dimensions(
+        tmp_local_netcdf, read_write_matrix[0], read_write_matrix[1]
+    )
+
+
+def test_no_circular_references(tmp_local_netcdf):
+    # https://github.com/h5py/h5py/issues/2019
+    with h5netcdf.File(tmp_local_netcdf, "w") as ds:
+        ds.dimensions["x"] = 2
+        ds.dimensions["y"] = 2
+
+    gc.collect()
+    with h5netcdf.File(tmp_local_netcdf, "r") as ds:
+        assert len(gc.get_referrers(ds)) == 1
+
+
+def test_creation_with_h5netcdf_edit_with_netcdf4(tmp_local_netcdf):
+    # In version 0.12.0, the wrong file creation attributes were used
+    # making netcdf4 unable to open files created by h5netcdf
+    # https://github.com/h5netcdf/h5netcdf/issues/128
+    with h5netcdf.File(tmp_local_netcdf, "w") as the_file:
+        the_file.dimensions = {"x": 5}
+        variable = the_file.create_variable("hello", ("x",), float)
+        variable[...] = 5
+
+    with netCDF4.Dataset(tmp_local_netcdf, mode="a") as the_file:
+        variable = the_file["hello"]
+        np.testing.assert_array_equal(variable[...].data, 5)
+        # Edit an existing variable
+        variable[:3] = 2
+
+        # Create a new variable
+        variable = the_file.createVariable("goodbye", float, ("x",))
+        variable[...] = 10
+
+    with h5netcdf.File(tmp_local_netcdf, "a") as the_file:
+        # Ensure edited variable is consistent with the expected data
+        variable = the_file["hello"]
+        np.testing.assert_array_equal(variable[...].data, [2, 2, 2, 5, 5])
+
+        # Ensure new variable is accessible
+        variable = the_file["goodbye"]
+        np.testing.assert_array_equal(variable[...].data, 10)
+
+
+def test_track_order_false(tmp_local_netcdf):
+    # track_order must be specified as True or not specified at all
+    # https://github.com/h5netcdf/h5netcdf/issues/130
+    with pytest.raises(ValueError):
+        h5netcdf.File(tmp_local_netcdf, "w", track_order=False)
+
+    with h5netcdf.File(tmp_local_netcdf, "w", track_order=True):
+        pass
+
+
+def test_monitor_netcdf4_requirement_for_track_order(tmp_local_netcdf):
+    # https://github.com/h5netcdf/h5netcdf/issues/130
+    with h5py.File(tmp_local_netcdf, "w", track_order=False):
+        pass
+    with pytest.raises(OSError):
+        with netCDF4.Dataset(tmp_local_netcdf, mode="a"):
+            pass
